package javagi.benchmarks.interpreter;

interface Eval {
    long eval(VarEnv env, FunEnv fenv);
}

implementation Eval[Expression] {
    long eval(VarEnv env, FunEnv fenv) {
        throw new RuntimeException("eval not implemented for " + this);
    }
}

implementation Eval[EqualsExpression] {
    public long eval(VarEnv env, FunEnv fenv) {
        long l = this.left.eval(env, fenv);
        long r = this.right.eval(env, fenv);
        if (l == r) {
            return 1;
        } else {
            return 0;
        }
    }
}

implementation Eval[MultExpression] {
    public long eval(VarEnv env, FunEnv fenv) {
        long l = this.left.eval(env, fenv);
        long r = this.right.eval(env, fenv);
        return l * r;
    }
}

implementation Eval[MinusExpression] {
    public long eval(VarEnv env, FunEnv fenv) {
        long l = this.left.eval(env, fenv);
        long r = this.right.eval(env, fenv);
        return l - r;
    }
}

implementation Eval[PlusExpression] {
    public long eval(VarEnv env, FunEnv fenv) {
        long l = this.left.eval(env, fenv);
        long r = this.right.eval(env, fenv);
        return l + r;
    }
}

implementation Eval[FuncallExpression] {
    public long eval(VarEnv env, FunEnv fenv) {
        Function fun = fenv.getFunction(this.name);
        long[] is = new long[this.args.length];
        for (int i = 0; i < is.length; i++) {
            is[i] = this.args[i].eval(env, fenv);
        }
        VarEnv env2 = fun.getVarEnv(is);
        return fun.getBody().eval(env2, fenv);
    }
}

implementation Eval[IfExpression] {
    public long eval(VarEnv env, FunEnv fenv) {
        long i = this.cond.eval(env, fenv);
        if (i != 0) {
            return this.then_.eval(env, fenv);
        } else {
            return this.else_.eval(env, fenv);
        }
    }
}

implementation Eval[VarExpression] {
    public long eval(VarEnv env, FunEnv fenv) {
        return env.getVar(this.name);
    }
}

implementation Eval[IntLiteral] {
    public long eval(VarEnv env, FunEnv fenv) {
        return this.value;
    }
}