import javagi.benchmarks.*;

public class Main {
    static int N = 10000000;

    final static void testVirtual(C c, int i, int j) {
        for (int k = 0; k < N; k++) {
            c.bar(i, j);
            c.bar(j, i);
            c.bar(i, j);
            c.bar(j, i);
            c.bar(i, j);
            c.bar(j, i);
            c.bar(i, j); // 7
        }
    }

    final static void testInterface(J c, int i, int j) {
        for (int k = 0; k < N; k++) {
            c.bar(i, j);
            c.bar(j, i);
            c.bar(i, j);
            c.bar(j, i);
            c.bar(i, j);
            c.bar(j, i);
            c.bar(i, j); // 7
        }
    }

    final static void testRetroactive(C c, int i, int j) {
        for (int k = 0; k < N; k++) {
            c.foo(i, j);
            c.foo(j, i);
            c.foo(i, j);
            c.foo(j, i);
            c.foo(i, j);
            c.foo(j, i);
            c.foo(i, j); // 7
        }
    }

    public static void main(String[] args) throws Exception {
        long start;
        long end;

	C c = new C();
	D d = new D();
	E e = new E();
        for (int run = 1; run <= Benchmarks.runCount(); run++) {
            System.out.println("testing invokevirtual");
            start = System.currentTimeMillis();
            testVirtual(c, 1, 0);
            testVirtual(d, 1, 0);
            testVirtual(e, 1, 0);
            end = System.currentTimeMillis();
            System.out.println("finished testing invokevirtual");
            Benchmarks.reportMethodCallResult(MethodCallKind.Virtual, run, end - start);
	}
	
	c.f = 0;
	d.f = 0;
	e.f = 0;
        for (int run = 1; run <= Benchmarks.runCount(); run++) {
            System.out.println("testing invokeinterface");
            start = System.currentTimeMillis();
            testInterface(c, 1, 0);
            testInterface(d, 1, 0);
            testInterface(e, 1, 0);
            end = System.currentTimeMillis();
            System.out.println("finished testing invokeinterface");
            Benchmarks.reportMethodCallResult(MethodCallKind.Interface, run, end - start);
	}

        for (int run = 1; run <= Benchmarks.runCount(); run++) {
            System.out.println("testing retroactive call");
            start = System.currentTimeMillis();
            testRetroactive(c, 1, 0);
            testRetroactive(d, 1, 0);
            testRetroactive(e, 1, 0);
            end = System.currentTimeMillis();
            System.out.println("finished testing retroactive call");
            Benchmarks.reportMethodCallResult(MethodCallKind.Retroactive, run, end - start);
	}

        for (int run = 1; run <= Benchmarks.runCount(); run++) {
            java.lang.reflect.Method bar = C.class.getDeclaredMethod("bar", Integer.TYPE, Integer.TYPE);
            System.out.println("testing call via reflection");
            start = System.currentTimeMillis();
            for (int i = 0; i < N/10; i++) {
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1);
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1);
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1);
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1);
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1); // 10
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1);
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1);
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1);
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1);
                bar.invoke(c, 1, 0);
                bar.invoke(c, 0, 1); // 20
                bar.invoke(c, 1, 0); // 21
            }
            end = System.currentTimeMillis();
            System.out.println("finished call via reflection");
            Benchmarks.reportMethodCallResult(MethodCallKind.Reflection, run, (end - start) * 10);
        }
    }
}

interface I {
    void foo(int i, int j);
}

interface J {
    void bar(int i, int j);
}

class C implements J {
    static int staticF = 0;
    int f = 0;
    int g = 0;
    static void staticBar(int i, int j) {
        //staticF += i + j;
    }
    public void bar(int i, int j) {
        //f += i + j;
    }
}

class D extends C {
    public void bar(int i, int j) {
        //f += i + j + 1;
    }
}

class E extends D {
    public void bar(int i, int j) {
        //f += i + j + 2;
    }
    public void foo(int i, int j) {
        //g += i + j + 2;
    }
}

implementation I[C] {
    void foo(int i, int j) {
        //g += i + j;
    }
}

implementation I[D] {
    void foo(int i, int j) {
        //g += i + j + 1;
    }
}