outer = 1000
inner = 1250
repeatFactor = 50

javagiHeader = "class B {} class C extends B {}  interface I {}  implementation I[C] {}\n"
javaHeader = "class B {} class C extends B implements I {}  interface I {}\n"

classIndices n = take n [1..]

castIndices n = concat $ take (n `div` repeatFactor) 
                              [take repeatFactor (repeat i) | i <- [1..]]

genClasses n =
    concatMap (\i -> "class C" ++ show i ++ " extends C {}\n") 
              (classIndices n)

genCasts1 var n =
    concatMap (\i -> var ++ " = (I) new C" ++ show i ++ "();\n")
              (castIndices n)

genCasts2 varObj varClass n =
    concatMap (\i -> varObj ++ " = new C" ++ show i ++ "();\n" ++
                     varClass ++ " = (C) " ++ varObj ++ ";\n")
              (castIndices n)

genInstanceof1 var n =
    concatMap (\i -> var ++ " = new C" ++ show i ++ "() instanceof I;\n")
              (castIndices n)

genInstanceof2 varObj varBool n =
    concatMap (\i -> varObj ++ " = new C" ++ show i ++ "();\n" ++
                     varBool ++ " = " ++ varObj ++ " instanceof C;\n")
              (castIndices n)

genIdentity varIface varBool n =
    concatMap (\i -> varIface ++ " = new C" ++ show i ++ "();\n" ++
                     varBool ++ " = " ++ varIface ++ " == new C" ++ 
                     show i ++ "();\n")
              (castIndices n)

genClassLoading n =
    concatMap (\i -> "Class.forName(\"C" ++ show i ++ "\");\n") (take n [1..])

forLoop n body = "for (int i = 0; i < " ++ show n ++ "; i++) {\n" ++
                 body ++ "}\n"

timing body = 
    "    long start, end;\n" ++
    "    start = System.currentTimeMillis();\n" ++
    body ++
    "    end = System.currentTimeMillis();\n" ++
    "    return (end-start);\n"

genTest name = 
    "class " ++ name ++ "{\n" ++
    "  static void loadClasses() throws ClassNotFoundException {\n" ++
    genClassLoading inner ++
    "  }\n" ++
    "  public static long testCast1() {\n" ++
    "    I iface;\n" ++
    timing (forLoop outer (genCasts1 "iface" inner)) ++
    "  }\n" ++
    "  public static long testCast2() {\n" ++
    "    Object obj;\n" ++
    "    C c;\n" ++
    timing (forLoop outer (genCasts2 "obj" "c" inner)) ++
    "  }\n" ++
    "  public static long testCast3() {\n" ++
    "    B b;\n" ++
    "    C c;\n" ++
    timing (forLoop outer (genCasts2 "b" "c" inner)) ++
    "  }\n" ++

    "  public static long testInstanceof1() {\n" ++
    "    boolean bool;\n" ++
    timing (forLoop outer (genInstanceof1 "bool" inner)) ++
    "  }\n" ++
    "  public static long testInstanceof2() {\n" ++
    "    Object obj;\n" ++
    "    boolean bool;\n" ++
    timing (forLoop outer (genInstanceof2 "obj" "bool" inner)) ++
    "  }\n" ++
    "  public static long testInstanceof3() {\n" ++
    "    B b;\n" ++
    "    boolean bool;\n" ++
    timing (forLoop outer (genInstanceof2 "b" "bool" inner)) ++
    "  }\n" ++

    "  public static long testIdentity1() {\n" ++
    "    Object obj;\n" ++
    "    boolean bool;\n" ++
    timing (forLoop (3*outer) (genIdentity "obj" "bool" inner)) ++
    "  }\n" ++
    "  public static long testIdentity2() {\n" ++
    "    C c;\n" ++
    "    boolean bool;\n" ++
    timing (forLoop (3*outer) (genIdentity "c" "bool" inner)) ++
    "  }\n" ++

    "}\n"

genJava = writeFile "JavaTest.java" (javaHeader ++ genClasses inner ++ 
                                         genTest "JavaTest")

genJavaGI = writeFile "JavaGITest.java" 
            (javagiHeader ++ genClasses inner ++
             genTest "JavaGITest")

main = do genJava
          genJavaGI